@@ -21,23 +21,34 @@
 )]
 
 mod config;
+mod generic_client;
 
 use std::{
     borrow::Cow,
     collections::HashMap,
+    fmt,
+    future::Future,
     ops::{Deref, DerefMut},
+    pin::Pin,
     sync::{
         atomic::{AtomicUsize, Ordering},
         Arc, Mutex, RwLock, Weak,
     },
 };
 
-use deadpool::{async_trait, managed};
+use deadpool::managed;
+#[cfg(not(target_arch = "wasm32"))]
 use tokio::spawn;
+use tokio::task::JoinHandle;
 use tokio_postgres::{
-    tls::MakeTlsConnect, tls::TlsConnect, types::Type, Client as PgClient, Config as PgConfig,
-    Error, IsolationLevel, Socket, Statement, Transaction as PgTransaction,
-    TransactionBuilder as PgTransactionBuilder,
+    types::Type, Client as PgClient, Config as PgConfig, Error, IsolationLevel, Statement,
+    Transaction as PgTransaction, TransactionBuilder as PgTransactionBuilder,
+};
+
+#[cfg(not(target_arch = "wasm32"))]
+use tokio_postgres::{
+    tls::{MakeTlsConnect, TlsConnect},
+    Socket,
 };
 
 pub use tokio_postgres;
@@ -47,25 +58,28 @@ pub use self::config::{
     TargetSessionAttrs,
 };
 
+pub use self::generic_client::GenericClient;
+
 pub use deadpool::managed::reexports::*;
 deadpool::managed_reexports!(
     "tokio_postgres",
     Manager,
-    deadpool::managed::Object<Manager>,
+    managed::Object<Manager>,
     Error,
     ConfigError
 );
 
+type BoxFuture<'a, T> = Pin<Box<dyn Future<Output = T> + Send + 'a>>;
+
 /// Type alias for [`Object`]
 pub type Client = Object;
 
-type RecycleResult = deadpool::managed::RecycleResult<Error>;
-type RecycleError = deadpool::managed::RecycleError<Error>;
+type RecycleResult = managed::RecycleResult<Error>;
+type RecycleError = managed::RecycleError<Error>;
 
 /// [`Manager`] for creating and recycling PostgreSQL connections.
 ///
 /// [`Manager`]: managed::Manager
-#[allow(missing_debug_implementations)] // due to `StatementCaches`
 pub struct Manager {
     config: ManagerConfig,
     pg_config: PgConfig,
@@ -75,6 +89,7 @@ pub struct Manager {
 }
 
 impl Manager {
+    #[cfg(not(target_arch = "wasm32"))]
     /// Creates a new [`Manager`] using the given [`tokio_postgres::Config`] and
     /// `tls` connector.
     pub fn new<T>(pg_config: tokio_postgres::Config, tls: T) -> Self
@@ -87,6 +102,7 @@ impl Manager {
         Self::from_config(pg_config, tls, ManagerConfig::default())
     }
 
+    #[cfg(not(target_arch = "wasm32"))]
     /// Create a new [`Manager`] using the given [`tokio_postgres::Config`], and
     /// `tls` connector and [`ManagerConfig`].
     pub fn from_config<T>(pg_config: tokio_postgres::Config, tls: T, config: ManagerConfig) -> Self
@@ -96,38 +112,58 @@ impl Manager {
         T::TlsConnect: Sync + Send,
         <T::TlsConnect as TlsConnect<Socket>>::Future: Send,
     {
+        Self::from_connect(pg_config, ConfigConnectImpl { tls }, config)
+    }
+
+    /// Create a new [`Manager`] using the given [`tokio_postgres::Config`], and
+    /// `connect` impl and [`ManagerConfig`].
+    pub fn from_connect(
+        pg_config: tokio_postgres::Config,
+        connect: impl Connect + 'static,
+        config: ManagerConfig,
+    ) -> Self {
         Self {
             config,
             pg_config,
-            connect: Box::new(ConnectImpl { tls }),
+            connect: Box::new(connect),
             statement_caches: StatementCaches::default(),
         }
     }
 }
 
-#[async_trait]
+impl fmt::Debug for Manager {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("Manager")
+            .field("config", &self.config)
+            .field("pg_config", &self.pg_config)
+            //.field("connect", &self.connect)
+            .field("statement_caches", &self.statement_caches)
+            .finish()
+    }
+}
+
 impl managed::Manager for Manager {
     type Type = ClientWrapper;
     type Error = Error;
 
     async fn create(&self) -> Result<ClientWrapper, Error> {
-        let client = self.connect.connect(&self.pg_config).await?;
-        let client_wrapper = ClientWrapper::new(client);
+        let (client, conn_task) = self.connect.connect(&self.pg_config).await?;
+        let client_wrapper = ClientWrapper::new(client, conn_task);
         self.statement_caches
             .attach(&client_wrapper.statement_cache);
         Ok(client_wrapper)
     }
 
-    async fn recycle(&self, client: &mut ClientWrapper) -> RecycleResult {
+    async fn recycle(&self, client: &mut ClientWrapper, _: &Metrics) -> RecycleResult {
         if client.is_closed() {
-            log::info!(target: "deadpool.postgres", "Connection could not be recycled: Connection closed");
-            return Err(RecycleError::StaticMessage("Connection closed"));
+            tracing::warn!(target: "deadpool.postgres", "Connection could not be recycled: Connection closed");
+            return Err(RecycleError::message("Connection closed"));
         }
         match self.config.recycling_method.query() {
             Some(sql) => match client.simple_query(sql).await {
                 Ok(_) => Ok(()),
                 Err(e) => {
-                    log::info!(target: "deadpool.postgres", "Connection could not be recycled: {}", e);
+                    tracing::warn!(target: "deadpool.postgres", "Connection could not be recycled: {}", e);
                     Err(e.into())
                 }
             },
@@ -140,44 +176,63 @@ impl managed::Manager for Manager {
     }
 }
 
-#[async_trait]
-trait Connect: Sync + Send {
-    async fn connect(&self, pg_config: &PgConfig) -> Result<PgClient, Error>;
+/// Describes a mechanism for establishing a connection to a PostgreSQL
+/// server via `tokio_postgres`.
+pub trait Connect: Sync + Send {
+    /// Establishes a new `tokio_postgres` connection, returning
+    /// the associated `Client` and a `JoinHandle` to a tokio task
+    /// for processing the connection.
+    fn connect(
+        &self,
+        pg_config: &PgConfig,
+    ) -> BoxFuture<'_, Result<(PgClient, JoinHandle<()>), Error>>;
 }
 
-struct ConnectImpl<T>
+#[cfg(not(target_arch = "wasm32"))]
+/// Provides an implementation of [`Connect`] that establishes the connection
+/// using the `tokio_postgres` configuration itself.
+#[derive(Debug)]
+pub struct ConfigConnectImpl<T>
 where
     T: MakeTlsConnect<Socket> + Clone + Sync + Send + 'static,
     T::Stream: Sync + Send,
     T::TlsConnect: Sync + Send,
     <T::TlsConnect as TlsConnect<Socket>>::Future: Send,
 {
-    tls: T,
+    /// The TLS connector to use for the connection.
+    pub tls: T,
 }
 
-#[async_trait]
-impl<T> Connect for ConnectImpl<T>
+#[cfg(not(target_arch = "wasm32"))]
+impl<T> Connect for ConfigConnectImpl<T>
 where
     T: MakeTlsConnect<Socket> + Clone + Sync + Send + 'static,
     T::Stream: Sync + Send,
     T::TlsConnect: Sync + Send,
     <T::TlsConnect as TlsConnect<Socket>>::Future: Send,
 {
-    async fn connect(&self, pg_config: &PgConfig) -> Result<PgClient, Error> {
-        let (client, connection) = pg_config.connect(self.tls.clone()).await?;
-        drop(spawn(async move {
-            if let Err(e) = connection.await {
-                log::warn!(target: "deadpool.postgres", "Connection error: {}", e);
-            }
-        }));
-        Ok(client)
+    fn connect(
+        &self,
+        pg_config: &PgConfig,
+    ) -> BoxFuture<'_, Result<(PgClient, JoinHandle<()>), Error>> {
+        let tls = self.tls.clone();
+        let pg_config = pg_config.clone();
+        Box::pin(async move {
+            let fut = pg_config.connect(tls);
+            let (client, connection) = fut.await?;
+            let conn_task = spawn(async move {
+                if let Err(e) = connection.await {
+                    tracing::warn!(target: "deadpool.postgres", "Connection error: {}", e);
+                }
+            });
+            Ok((client, conn_task))
+        })
     }
 }
 
 /// Structure holding a reference to all [`StatementCache`]s and providing
 /// access for clearing all caches and removing single statements from them.
-#[derive(Default)]
-#[allow(missing_debug_implementations)] // due to `StatementCache`
+#[derive(Default, Debug)]
 pub struct StatementCaches {
     caches: Mutex<Vec<Weak<StatementCache>>>,
 }
@@ -216,6 +271,15 @@ impl StatementCaches {
     }
 }
 
+impl fmt::Debug for StatementCache {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("ClientWrapper")
+            //.field("map", &self.map)
+            .field("size", &self.size)
+            .finish()
+    }
+}
+
 // Allows us to use owned keys in a `HashMap`, but still be able to call `get`
 // with borrowed keys instead of allocating them each time.
 #[derive(Debug, Eq, Hash, PartialEq)]
@@ -243,7 +307,6 @@ struct StatementCacheKey<'a> {
 /// Normally, you probably want to use the [`ClientWrapper::prepare_cached()`]
 /// and [`ClientWrapper::prepare_typed_cached()`] methods instead (or the
 /// similar ones on [`Transaction`]).
-#[allow(missing_debug_implementations)] // due to `Statement`
 pub struct StatementCache {
     map: RwLock<HashMap<StatementCacheKey<'static>, Statement>>,
     size: AtomicUsize,
@@ -343,33 +406,38 @@ impl StatementCache {
 }
 
 /// Wrapper around [`tokio_postgres::Client`] with a [`StatementCache`].
-#[allow(missing_debug_implementations)] // due to `StatementCache`
+#[derive(Debug)]
 pub struct ClientWrapper {
     /// Original [`PgClient`].
     client: PgClient,
 
+    /// A handle to the connection task that should be aborted when the client
+    /// wrapper is dropped.
+    conn_task: JoinHandle<()>,
+
     /// [`StatementCache`] of this client.
     pub statement_cache: Arc<StatementCache>,
 }
 
 impl ClientWrapper {
     /// Create a new [`ClientWrapper`] instance using the given
-    /// [`tokio_postgres::Client`].
+    /// [`tokio_postgres::Client`] and handle to the connection task.
     #[must_use]
-    pub fn new(client: PgClient) -> Self {
+    pub fn new(client: PgClient, conn_task: JoinHandle<()>) -> Self {
         Self {
             client,
+            conn_task,
             statement_cache: Arc::new(StatementCache::new()),
         }
     }
 
-    /// Like [`tokio_postgres::Transaction::prepare()`], but uses an existing
+    /// Like [`tokio_postgres::Client::prepare()`], but uses an existing
     /// [`Statement`] from the [`StatementCache`] if possible.
     pub async fn prepare_cached(&self, query: &str) -> Result<Statement, Error> {
         self.statement_cache.prepare(&self.client, query).await
     }
 
-    /// Like [`tokio_postgres::Transaction::prepare_typed()`], but uses an
+    /// Like [`tokio_postgres::Client::prepare_typed()`], but uses an
     /// existing [`Statement`] from the [`StatementCache`] if possible.
     pub async fn prepare_typed_cached(
         &self,
@@ -415,15 +483,20 @@ impl DerefMut for ClientWrapper {
     }
 }
 
+impl Drop for ClientWrapper {
+    fn drop(&mut self) {
+        self.conn_task.abort()
+    }
+}
+
 /// Wrapper around [`tokio_postgres::Transaction`] with a [`StatementCache`]
 /// from the [`Client`] object it was created by.
-#[allow(missing_debug_implementations)] // due to `StatementCache`
 pub struct Transaction<'a> {
     /// Original [`PgTransaction`].
     txn: PgTransaction<'a>,
 
     /// [`StatementCache`] of this [`Transaction`].
-    statement_cache: Arc<StatementCache>,
+    pub statement_cache: Arc<StatementCache>,
 }
 
 impl<'a> Transaction<'a> {
@@ -479,6 +552,15 @@ impl<'a> Transaction<'a> {
     }
 }
 
+impl<'a> fmt::Debug for Transaction<'a> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("Transaction")
+            //.field("txn", &self.txn)
+            .field("statement_cache", &self.statement_cache)
+            .finish()
+    }
+}
+
 impl<'a> Deref for Transaction<'a> {
     type Target = PgTransaction<'a>;
 
@@ -495,7 +577,6 @@ impl<'a> DerefMut for Transaction<'a> {
 
 /// Wrapper around [`tokio_postgres::TransactionBuilder`] with a
 /// [`StatementCache`] from the [`Client`] object it was created by.
-#[allow(missing_debug_implementations)] // due to `StatementCache`
 #[must_use = "builder does nothing itself, use `.start()` to use it"]
 pub struct TransactionBuilder<'a> {
     /// Original [`PgTransactionBuilder`].
@@ -555,6 +636,15 @@ impl<'a> TransactionBuilder<'a> {
     }
 }
 
+impl<'a> fmt::Debug for TransactionBuilder<'a> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("TransactionBuilder")
+            //.field("builder", &self.builder)
+            .field("statement_cache", &self.statement_cache)
+            .finish()
+    }
+}
+
 impl<'a> Deref for TransactionBuilder<'a> {
     type Target = PgTransactionBuilder<'a>;
 
